// Fix compiler bug
#define for if(0) ; else for

#include <Windows.h>
#include <stdio.h>

#include "..\AVISynthSDK\avisynth.h"

typedef unsigned char byte;

class Filter : public GenericVideoFilter
{
private:
  bool   m_fDoubleRate;
  bool   m_fBlend;
  bool   m_fShowDeinterlaced;
  bool   m_fLog;
  int    m_iLaceThresh;
  int    m_iEdgeThresh;
  int    m_iInterframeLeaveThresh;
  int    m_iInterframeAvg;
  float  m_rLogLacePercent;
  int*   m_piFrameDiffs;
  HANDLE m_hLog;
  bool   m_fFieldBasedInput;
  PClip  m_pChildWeave;
  PClip  m_pChildDoubleWeave;
  const char* m_szLogFile;

public:
  Filter(PClip pChildNormal,
         PClip pChildWeave,
         PClip pChildDoubleWeave,
         int   iLaceThresh,
         int   iEdgeThresh,
         int   iInterframeLeaveThresh,
         int   iInterframeAvg,
         bool  fDoubleRate,
         bool  fBlend,
         bool  fShowDeinterlaced,
         bool  fLog,
         float rLogLacePercent,
         const char* szLogFile) :
   // All three types of input streams are stored.  Which one is used
   // is determined below.
   GenericVideoFilter(pChildNormal),
   m_pChildWeave(pChildWeave),
   m_pChildDoubleWeave(pChildDoubleWeave),
   m_fDoubleRate(fDoubleRate),
   m_fBlend(fBlend),
   m_fShowDeinterlaced(fShowDeinterlaced),
   m_fLog(fLog),
   m_iLaceThresh(iLaceThresh),
   m_iEdgeThresh(iEdgeThresh),
   m_iInterframeLeaveThresh(iInterframeLeaveThresh),
   m_iInterframeAvg(iInterframeAvg),
   m_rLogLacePercent(rLogLacePercent),
   m_szLogFile(szLogFile),
   m_hLog(INVALID_HANDLE_VALUE),
   m_fFieldBasedInput(vi.field_based)
  {
/*
{
  char sz[512];
  sprintf(sz, "tff=%d, doublerate=%d, blend=%d, showlace=%d, log=%d,\n"
              "lacethresh=%d, edgethresh=%d, staticthresh=%d,\n"
              "staticavg=%d, logpercent=%f,\n"
              "width=%d, height=%d, fps_numerator=%d, fps_denominator=%d,\n"
              "num_frames=%d, pixel_type=0x%02X, field_based=%d,\n"
              "audio_samples_per_second=%d, num_audio_samples=%d,\n"
              "stereo=%d, sixteen_bit=%d",
          pChildDoubleWeave->GetParity(), m_fDoubleRate, m_fBlend, m_fShowDeinterlaced, m_fLog,
          m_iLaceThresh, m_iEdgeThresh, m_iInterframeLeaveThresh,
          m_iInterframeAvg, m_rLogLacePercent,
          vi.width, vi.height, vi.fps_numerator, vi.fps_denominator,
          vi.num_frames, vi.pixel_type, vi.field_based,
          vi.audio_samples_per_second, vi.num_audio_samples,
          vi.stereo, vi.sixteen_bit);
  MessageBox(NULL, sz, "INFO", MB_OK);
}
*/
    if(m_fFieldBasedInput)
    {
      // Input is fields, output is frames, which are twice as tall.
      vi.height *= 2;
      if(m_fDoubleRate)
      {
        // When the fields are collapsed to frames, the frame rate must
        // be halved.  But when we DoubleWeave() we must double the frame
        // rate.  The result is a wash, so we do nothing in this case.
      }
      else
      {
        // When the fields are collapsed to frames, we also halve the
        // frame rate and number of frames.
        vi.fps_denominator *= 2;
        vi.num_frames /= 2;
      }
    }
    else
    {
      if(m_fDoubleRate)
      {
        // When we DoubleWeave(), we need to double the frame rate and
        // number of frames.
        vi.fps_numerator *= 2;
        vi.num_frames *= 2;
      }
      else
      {
        // Frames in, frames out, means no change
      }
    }

    // We always output frames
    vi.field_based = false;

    // Get a place to put the interframe differences

    m_piFrameDiffs = new int[vi.width * vi.height];
    for(int i = 0; i < (vi.width * vi.height); i++)
    {
      m_piFrameDiffs[i] = (m_iInterframeLeaveThresh *
                           m_iInterframeLeaveThresh) + 1;
    }

    // Open the log file if necessary
    if(m_fLog)
    {
      m_hLog = CreateFile(m_szLogFile,
                          GENERIC_WRITE,
                          FILE_SHARE_READ,
                          NULL,
                          CREATE_ALWAYS,
                          FILE_ATTRIBUTE_NORMAL,
                          NULL);
      if(m_hLog != INVALID_HANDLE_VALUE)
      {
        DWORD dw;
        WriteFile(m_hLog,
                  "Frame\tInterlaced area (%)\r\n",
                  27,
                  &dw,
                  NULL);
      }
    }
  }
  virtual __stdcall ~Filter()
  {
    if(m_hLog != INVALID_HANDLE_VALUE)
    {
      CloseHandle(m_hLog);
    }
    delete [] m_piFrameDiffs;
  }
  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* pEnv);
};


PVideoFrame __stdcall Filter::GetFrame(int                 n,
                                       IScriptEnvironment* pEnv)
{
  PVideoFrame pvfCurrent;
  PVideoFrame pvfPrev;
  PVideoFrame pvfNew;

  int iDeinterlacedPixels = 0;

  // Determine which of the potential input streams is needed.
  PClip pChild;
  if(m_fDoubleRate)
    // Output is double the input.  The DoubleWeave() stream will
    // take the input (either fields or frames) and create twice
    // as many full frames, in which each field is associated with
    // the field before it and the field after it.
    pChild = m_pChildDoubleWeave;
  else
    if(m_fFieldBasedInput)
      // The input is in fields, but we want frames for our
      // input, so we choose the Weave() stream.
      pChild = m_pChildWeave;
    else
      // Frames in, frames out, so use the normal stream.
      pChild = child;

  if(m_fDoubleRate)
  {
    // A trick here to get the frames aligned correctly.
    if(n)
      --n;
  }
  pvfCurrent = pChild->GetFrame(n, pEnv);

  if(n > 1)
    pvfPrev = pChild->GetFrame(n - 2, pEnv);
  else
    pvfPrev = pvfCurrent;

  int iThreshold = m_iLaceThresh;
  iThreshold = iThreshold * iThreshold * 4;
  int iEdgeDetect = m_iEdgeThresh;
  if(iEdgeDetect > 180)
    iEdgeDetect = 180;  // We don't want an integer overflow in the interlace calculation.
  iEdgeDetect = iEdgeDetect * iEdgeDetect;
  bool bShowDeinterlacedAreaOnly = m_fShowDeinterlaced;
  bool bBlend = m_fBlend;
  int iInterFrameLeaveThreshold = m_iInterframeLeaveThresh *
                                  m_iInterframeLeaveThresh;
  int iInterFrameAverage = m_iInterframeAvg * 1024 / 100;
  if(iInterFrameAverage < 0)
    iInterFrameAverage = 0;
  if(iInterFrameAverage > 1023)
    iInterFrameAverage = 1023;
  int iInterFrameAverageRest = 1024 - iInterFrameAverage;

  const int iSrcPitch = pvfCurrent->GetPitch();

  if(vi.IsRGB32())
  {
    typedef unsigned long Pixel32;

    int iR0, iG0, iB0, iR1, iG1, iB1, iR2, iG2, iB2, iR3, iG3, iB3;
    int iR2p, iG2p, iB2p;
    Pixel32* pframe1;
    Pixel32* pframe2;
    Pixel32* pframe3;
    Pixel32* pdst1;
    Pixel32* pframe2p;
    int* piFrameDiffs2p;
    int iInterlaceValue0, iInterlaceValue1, iInterlaceValue2;
    int iFrameDiffValue0, iFrameDiffValue1, iFrameDiffValue2;
    iR1 = iG1 = iB1 = 0;  // Avoid compiler warning. The value is not used.

    Pixel32* pframe = (Pixel32*)pvfCurrent->GetReadPtr();
    Pixel32* pframeprev = (Pixel32*)pvfPrev->GetReadPtr();
    pvfNew = pEnv->NewVideoFrame(vi);
    Pixel32* pdst = (Pixel32*)pvfNew->GetWritePtr();
    int* piFrameDiffs = m_piFrameDiffs;

    const int iDstPitch = pvfNew->GetPitch();

    int iOddEven;
    if(m_fDoubleRate)
      iOddEven = (n & 1);
    else
      iOddEven = 1;

    // Do some area based deinterlacing on the combined frame.
    for(int x = 0; x < vi.width; x++)
    {
      pframe3 = pframe + x;
      iR3 = (*pframe3 >> 16) & 0xff;
      iG3 = (*pframe3 >> 8) & 0xff;
      iB3 = *pframe3 & 0xff;
      pframe2 = (Pixel32 *)((char *)pframe3 + iSrcPitch);
      iR2 = (*pframe2 >> 16) & 0xff;
      iG2 = (*pframe2 >> 8) & 0xff;
      iB2 = *pframe2 & 0xff;
    
      pframe2p = pframeprev + x;
      piFrameDiffs2p = piFrameDiffs + (x * vi.height);

      pdst1 = pdst + x;
      iInterlaceValue1 = iInterlaceValue2 = 0;
      iFrameDiffValue1 = iFrameDiffValue2 = 0;

      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
      for(int y = 0; y <= vi.height; y++)
      {
        pframe1 = pframe2;
        pframe2 = pframe3;
        pframe3 = (Pixel32 *)((char *)pframe3 + iSrcPitch);
        iR0 = iR1;
        iG0 = iG1;
        iB0 = iB1;
        iR1 = iR2;
        iG1 = iG2;
        iB1 = iB2;
        iR2 = iR3;
        iG2 = iG3;
        iB2 = iB3;
        if(y < vi.height - 1)
        {
          iR3 = (*pframe3 >> 16) & 0xff;
          iG3 = (*pframe3 >> 8) & 0xff;
          iB3 = *pframe3 & 0xff;

          iR2p = (*pframe2p >> 16) & 0xff;
          iG2p = (*pframe2p >> 8) & 0xff;
          iB2p = *pframe2p & 0xff;
        }
        else
        {
          iR3 = iR1;
          iG3 = iG1;
          iB3 = iB1;
        }
        pframe2p = (Pixel32 *)((char *)pframe2p + iSrcPitch);

        iInterlaceValue0 = iInterlaceValue1;
        iInterlaceValue1 = iInterlaceValue2;
        iFrameDiffValue0 = iFrameDiffValue1;
        iFrameDiffValue1 = iFrameDiffValue2;
        if(y < vi.height)
        {
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
          // 3, 6 and 1 are approximate values for converting R, G, B to intensity.
          iInterlaceValue2 = (3 * ((iR1 - iR2) * (iR3 - iR2) - ((iEdgeDetect * (iR1 - iR3) * (iR1 - iR3)) >> 12)) +
                    6 * ((iG1 - iG2) * (iG3 - iG2) - ((iEdgeDetect * (iG1 - iG3) * (iG1 - iG3)) >> 12)) +
                    (iB1 - iB2) * (iB3 - iB2) - ((iEdgeDetect * (iB1 - iB3) * (iB1 - iB3)) >> 12));
          iFrameDiffValue2 = (iInterFrameAverage * *piFrameDiffs2p +
                        iInterFrameAverageRest * (n > 1 ?
                          3 * (iR2 - iR2p) * (iR2 - iR2p) + 6 * (iG2 - iG2p) * (iG2 - iG2p) + (iB2 - iB2p) * (iB2 - iB2p) :
                          *piFrameDiffs2p)) >> 10;
          *piFrameDiffs2p = iFrameDiffValue2;
          piFrameDiffs2p++;
        }
        else
        {
          iInterlaceValue2 = 0;
          iFrameDiffValue2 = 0;
        }

        if(y > 0)
        {  // Get mean interlace value of 3 rows (i.e. 5 pixels are examined).
          // Middle row has twice the weight.
          if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold &&
            iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
          {
            if(bShowDeinterlacedAreaOnly)
            {
              *pdst1 = 0x00903030 + (bBlend ? (((iR0 + 2 * iR1 + iR2) >> 4) << 16) + (((iG0 + 2 * iG1 + iG2) >> 4) << 8) + ((iB0 + 2 * iB1 + iB2) >> 4)
                                  :
                                    (y % 2 == iOddEven ? (iR1 >> 2 << 16) + (iG1 >> 2 << 8) + (iB1 >> 2) : (((iR0 + iR2) >> 3) << 16) + (((iG0 + iG2) >> 3) << 8) + ((iB0 + iB2) >> 3)));
            }
            else
            {
              // Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
              // Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
              *pdst1 = bBlend ? (((iR0 + 2 * iR1 + iR2) >> 2) << 16) + (((iG0 + 2 * iG1 + iG2) >> 2) << 8) + ((iB0 + 2 * iB1 + iB2) >> 2)
                        :
                        (y % 2 == iOddEven ? *pframe1 : (((iR0 + iR2) >> 1) << 16) + (((iG0 + iG2) >> 1) << 8) + ((iB0 + iB2) >> 1));

            }
            iDeinterlacedPixels++;
          }
          else
          {
            if(bShowDeinterlacedAreaOnly)
            {
              *pdst1 = (iR1 >> 2 << 16) + (iG1 >> 2 << 8) + (iB1 >> 2);
              if(iInterlaceValue0 + 2 * iInterlaceValue1 + iInterlaceValue2 > iThreshold)
                *pdst1 += 0x00309030;
              else if(iFrameDiffValue0 + 2 * iFrameDiffValue1 + iFrameDiffValue2 >= iInterFrameLeaveThreshold)
                *pdst1 += 0x00303090;
              else
                *pdst1 += 0x00303030;
            }
            else
              *pdst1 = *pframe1;
          }
          pdst1 = (Pixel32 *)((char *)pdst1 + iDstPitch);
        }
      }
    }
  }
  else if(vi.IsYUY2())
  {
    struct PixelPair
    {
      BYTE Y0;
      BYTE U;
      BYTE Y1;
      BYTE V;
    };

    PixelPair PixelPair0, PixelPair1 = {0, 0, 0, 0}, PixelPair2, PixelPair3;
    int iY2pe, iY2po;
    PixelPair *pframe1, *pframe2, *pframe3, *pdst1, *pframe2p;
    int *piFrameDiffs2p;
    int iInterlaceValue0e, iInterlaceValue1e, iInterlaceValue2e;
    int iInterlaceValue0o, iInterlaceValue1o, iInterlaceValue2o;
    int iFrameDiffValue0e, iFrameDiffValue1e, iFrameDiffValue2e;
    int iFrameDiffValue0o, iFrameDiffValue1o, iFrameDiffValue2o;

    PixelPair* pframe = (PixelPair*)pvfCurrent->GetReadPtr();
    PixelPair* pframeprev = (PixelPair*)pvfPrev->GetReadPtr();
    pvfNew = pEnv->NewVideoFrame(vi);
    PixelPair* pdst = (PixelPair*)pvfNew->GetWritePtr();
    int* piFrameDiffs = m_piFrameDiffs;

    const int iDstPitch = pvfNew->GetPitch();

    int iOddEven;
    if(m_fDoubleRate)
      iOddEven = (n & 1);
    else
      iOddEven = 1;

    // Do some area based deinterlacing on the combined frame.
    for(int x = 0; x < vi.width; x += 2)
    {
      pframe3 = pframe + (x >> 1);
      PixelPair3 = *pframe3;
      pframe2 = (PixelPair*)((DWORD)pframe3 + iSrcPitch);
      PixelPair2 = *pframe2;
    
      pframe2p = pframeprev + (x >> 1);
      piFrameDiffs2p = piFrameDiffs + (x * vi.height);

      pdst1 = pdst + (x >> 1);
      iInterlaceValue1e = iInterlaceValue2e = 0;
      iInterlaceValue1o = iInterlaceValue2o = 0;
      iFrameDiffValue1e = iFrameDiffValue2e = 0;
      iFrameDiffValue1o = iFrameDiffValue2o = 0;

      // In this loop, output is delayed one loop (or row) from input. Current output row has variables with digit 1 and current input row has digit 2.
      for(int y = 0; y <= vi.height; y++)
      {
        pframe1 = pframe2;
        pframe2 = pframe3;
        pframe3 = (PixelPair*)((DWORD)pframe3 + iSrcPitch);
        PixelPair0 = PixelPair1;
        PixelPair1 = PixelPair2;
        PixelPair2 = PixelPair3;
        if(y < vi.height - 1)
        {
          PixelPair3 = *pframe3;
          iY2pe = *((unsigned char*)pframe2p);
          iY2po = *((unsigned char*)pframe2p + 2);
        }
        else
        {
          PixelPair3 = PixelPair1;
        }
        pframe2p = (PixelPair*)((DWORD)pframe2p + iSrcPitch);

        iInterlaceValue0e = iInterlaceValue1e;
        iInterlaceValue0o = iInterlaceValue1o;
        iInterlaceValue1e = iInterlaceValue2e;
        iInterlaceValue1o = iInterlaceValue2o;
        iFrameDiffValue0e = iFrameDiffValue1e;
        iFrameDiffValue0o = iFrameDiffValue1o;
        iFrameDiffValue1e = iFrameDiffValue2e;
        iFrameDiffValue1o = iFrameDiffValue2o;
        if(y < vi.height)
        {
          // Calculate the interlace value by checking if the pixel color on previous row differs much
          // from this row and next row differs much (with same sign) from this row.
          // Detect edges by checking so that pixel color on previous row doesn't differ too much from next row.
          // If it does, it's probably just an ordinary edge.
          iInterlaceValue2e = 10 * ((PixelPair1.Y0 - PixelPair2.Y0) * (PixelPair3.Y0 - PixelPair2.Y0) - ((iEdgeDetect * (PixelPair1.Y0 - PixelPair3.Y0) * (PixelPair1.Y0 - PixelPair3.Y0)) >> 12));
          iInterlaceValue2o = 10 * ((PixelPair1.Y1 - PixelPair2.Y1) * (PixelPair3.Y1 - PixelPair2.Y1) - ((iEdgeDetect * (PixelPair1.Y1 - PixelPair3.Y1) * (PixelPair1.Y1 - PixelPair3.Y1)) >> 12));
          iFrameDiffValue2e = (iInterFrameAverage * *piFrameDiffs2p + iInterFrameAverageRest *
                              (n                       ?
                               10 * (PixelPair2.Y0 - iY2pe) * (PixelPair2.Y0 - iY2pe) :
                               *piFrameDiffs2p)) >> 10;
          iFrameDiffValue2o = (iInterFrameAverage * *(piFrameDiffs2p + 1) + iInterFrameAverageRest *
                              (n                       ?
                               10 * (PixelPair2.Y1 - iY2po) * (PixelPair2.Y1 - iY2po) :
                               *(piFrameDiffs2p + 1))) >> 10;
          *piFrameDiffs2p = iFrameDiffValue2e;
          *(piFrameDiffs2p + 1) = iFrameDiffValue2o;
          piFrameDiffs2p++;
        }
        else
        {
          iInterlaceValue2e = 0;
          iInterlaceValue2o = 0;
          iFrameDiffValue2e = 0;
          iFrameDiffValue2o = 0;
        }

        if(y > 0)
        {
          // Get mean interlace value of 3 rows (i.e. 5 pixels are examined).
          // Middle row has twice the weight.
          if(iInterlaceValue0e + 2 * iInterlaceValue1e + iInterlaceValue2e > iThreshold &&
             iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
          {
            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold &&
               iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
            {
              // Even and odd are both interlaced
              if(bBlend)
              {
                // Blend: Get mean value of previous and next row (weight 0.25) and this row (weight 0.5).
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = (PixelPair0.Y0 + 2 * PixelPair1.Y0 + PixelPair2.Y0) / 4;
                PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U  + PixelPair2.U)  / 4;
                PixelPairDst.Y1 = (PixelPair0.Y1 + 2 * PixelPair1.Y1 + PixelPair2.Y1) / 4;
                PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V  + PixelPair2.V)  / 4;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                // Interpolate: Odd lines: Copy from source. Even lines: Get mean value of previous and next row
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = (PixelPair0.Y0 + PixelPair2.Y0) / 2;
                  PixelPairDst.U  = (PixelPair0.U  + PixelPair2.U)  / 2;
                  PixelPairDst.Y1 = (PixelPair0.Y1 + PixelPair2.Y1) / 2;
                  PixelPairDst.V  = (PixelPair0.V  + PixelPair2.V)  / 2;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels += 2;
            }
            else
            {
              // Even is interlaced, but odd is not
              if(bBlend)
              {
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = (PixelPair0.Y0 + 2 * PixelPair1.Y0 + PixelPair2.Y0) / 4;
                PixelPairDst.U  = (PixelPair0.U  + 6 * PixelPair1.U  + PixelPair2.U)  / 8;
                PixelPairDst.Y1 = PixelPair1.Y1;
                PixelPairDst.V  = (PixelPair0.V  + 6 * PixelPair1.V  + PixelPair2.V)  / 8;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = (PixelPair0.Y0 + PixelPair2.Y0) / 2;
                  PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U + PixelPair2.U)  / 4;
                  PixelPairDst.Y1 = PixelPair1.Y1;
                  PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V + PixelPair2.V)  / 4;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels++;
            }
          }
          else
          {
            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold &&
               iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
            {
              // Even is not interlaced, but odd is
              if(bBlend)
              {
                PixelPair PixelPairDst;
                PixelPairDst.Y0 = PixelPair1.Y0;
                PixelPairDst.U  = (PixelPair0.U  + 6 * PixelPair1.U  + PixelPair2.U)  / 8;
                PixelPairDst.Y1 = (PixelPair0.Y1 + 2 * PixelPair1.Y1 + PixelPair2.Y1) / 4;
                PixelPairDst.V  = (PixelPair0.V  + 6 * PixelPair1.V  + PixelPair2.V)  / 8;
                *pdst1 = PixelPairDst;
              }
              else // Interpolate
              {
                if((y & 1) ^ iOddEven)
                {
                  *pdst1 = *pframe1;
                }
                else
                {
                  PixelPair PixelPairDst;
                  PixelPairDst.Y0 = PixelPair1.Y0;
                  PixelPairDst.U  = (PixelPair0.U  + 2 * PixelPair1.U + PixelPair2.U)  / 4;
                  PixelPairDst.Y1 = (PixelPair0.Y1 + PixelPair2.Y1) / 2;
                  PixelPairDst.V  = (PixelPair0.V  + 2 * PixelPair1.V + PixelPair2.V)  / 4;
                  *pdst1 = PixelPairDst;
                }
              }

              iDeinterlacedPixels++;
            }
            else
            {
              // Neither even nor odd is interlaced
              *pdst1 = *pframe1;
            }
          }

          if(bShowDeinterlacedAreaOnly)
          {
            int iY0 = pdst1->Y0 / 4;
            int iU0 = (pdst1->U - 128) / 4;
            int iU1 = iU0;
            int iY1 = pdst1->Y1 / 4;
            int iV0 = (pdst1->V - 128) / 4;
            int iV1 = iV0;

            if(iInterlaceValue0e + 2 * iInterlaceValue1e + iInterlaceValue2e > iThreshold)
            {
              if(iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
              {
                iY0 += 76;
                iU0 += -14;
                iV0 += 41;
              }
              else
              {
                iY0 += 103;
                iU0 += -28;
                iV0 += -34;
              }
            }
            else
            {
              if(iFrameDiffValue0e + 2 * iFrameDiffValue1e + iFrameDiffValue2e >= iInterFrameLeaveThreshold)
              {
                iY0 += 58;
                iU0 += 41;
                iV0 += -6;
              }
              else
              {
                iY0 += 47;
              }
            }

            if(iInterlaceValue0o + 2 * iInterlaceValue1o + iInterlaceValue2o > iThreshold)
            {
              if(iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
              {
                iY1 += 76;
                iU1 += -14;
                iV1 += 41;
              }
              else
              {
                iY1 += 103;
                iU1 += -28;
                iV1 += -34;
              }
            }
            else
            {
              if(iFrameDiffValue0o + 2 * iFrameDiffValue1o + iFrameDiffValue2o >= iInterFrameLeaveThreshold)
              {
                iY1 += 58;
                iU1 += 41;
                iV1 += -6;
              }
              else
              {
                iY1 += 47;
              }
            }

            pdst1->Y0 = iY0;
            pdst1->U = ((iU0 + iU1) / 2) + 128;
            pdst1->Y1 = iY1;
            pdst1->V = ((iV0 + iV1) / 2) + 128;
          }

          pdst1 = (PixelPair*)((DWORD)pdst1 + iDstPitch);
        }
      }
    }
  }
  else
  {
    pEnv->ThrowError("SmoothDeinterlace only works on RGB32 and YUY2 formats.");
    pvfNew = pvfCurrent;
  }

  if(m_fLog)
  {
    float fDeinterlacedPercent = 100.0f * iDeinterlacedPixels / vi.width / vi.height;
    if(fDeinterlacedPercent > m_rLogLacePercent && m_hLog != INVALID_HANDLE_VALUE)
    {
      char szLog[25];
      sprintf(szLog, "%li\t%5.1f\r\n", n, fDeinterlacedPercent);
      DWORD dw;
      WriteFile(m_hLog, szLog, strlen(szLog), &dw, NULL);
    }
  }

  return(pvfNew);
}

AVSValue __cdecl Create_Filter(AVSValue            args,
                               void*               user_data,
                               IScriptEnvironment* pEnv)
{
  // The input could be separate fields or interlaced frames, and the
  // user could want to double the framerate or not.  To accomodate
  // all four of the possible combinations, we would need one of four
  // different types of input streams:  normal, Weave,
  // DoubleWeaveFields, and DoubleWeaveFrames.  [DoubleWeave()
  // handles differentiating between DoubleWeaveFields() and
  // DoubleFieldFrames(), so we only have to deal with a single
  // DoubleWeave() stream.]  The table below shows the correct
  // process for the two input formats and the two desired output
  // formats:
  //
  //           Display Rate (doublerate=)
  //        Frame (false)      Field (true)
  // Fields    Weave       DoubleWeave(Fields)
  // Frames     None       DoubleWeave(Frames)
  
  // Prepare all three types of input streams (only one will be used).
  PClip pChildNormal = args[0].AsClip();
  PClip pChildWeave = pEnv->Invoke("Weave", args[0]).AsClip();
  AVSValue avs;
  if(args[5].Defined())
    if(args[5].AsBool())
      avs = pEnv->Invoke("AssumeTFF", args[0]);
    else
      avs = pEnv->Invoke("AssumeBFF", args[0]);
  else
    avs = args[0];
  PClip pChildDoubleWeave = pEnv->Invoke("DoubleWeave", avs).AsClip();

  return(new Filter(pChildNormal,
                    pChildWeave,
                    pChildDoubleWeave,
                    args[1].AsInt(24),
                    args[2].AsInt(20),
                    args[3].AsInt(35),
                    args[4].AsInt(80),
                    args[6].AsBool(false),
                    args[7].AsBool(false),
                    args[8].AsBool(false),
                    args[9].AsBool(false),
                    (float)args[10].AsFloat(0.0),
                    args[11].AsString("DeinterlaceSmooth.log")));
}

extern "C"
__declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* pEnv)
{
  pEnv->AddFunction("SmoothDeinterlace", "c[lacethresh]i[edgethresh]i[staticthresh]i[staticavg]i[tff]b[doublerate]b[blend]b[showlace]b[log]b[logpercent]f[logfile]s", Create_Filter, 0);
  return("`SmoothDeinterlace' plugin 1.4");
}
